/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Action
Version:   1.1
Status:    Stable
ID:        ID_ACTION
Date:      March 2003
Category:  Command
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 1999-2003.  All rights reserved.
Short:     The action class is used to perform any action on a shared object.
Keywords:  action, object, monitor, function, call

###DESCRIPTION###
<p>The Action class allows you to execute actions and methods on other
objects in the system.  It is specifically designed for use in scripts, such as
the Dynamic Markup Language.  The Action class
is especially useful for executing generic actions such as Free,
Show, MoveToFront and Resize during the execution of a script. You may
also create static Action objects, which are useful for reacting to the
user's activity and other events in the system.  Another useful feature
is action monitoring, which allows you to react to action calls made on
any object.</p>

<p>Executing an action requires that you state the name of the
action that you want to call and the object that is to receive it.  Following
that, any arguments relevant to the type of action should be specified.  It
is vital that you set all of the arguments related to the action, e.g. the
Resize action requires that three arguments are set.  If you do not set all
of the required arguments, the default settings will be used, which may give
unexpected results.  Here are some DML based examples:</p>

<pre>
  &lt;action call=&quot;Free&quot; object=&quot;[myobject]&quot;/&gt;
  &lt;action call=&quot;Resize&quot; object=&quot;[renderobject]&quot; &width=&quot;100&quot; &height=&quot;150&quot; &depth=&quot;0&quot;/&gt;
  &lt;action static method=&quot;List&quot; object=&quot;[container]&quot;/&gt;
</pre>

<p>This next example monitors an object and performs a reaction whenever it
is hidden.  Note:  When using action monitoring, the Action object will
automatically be made static.</p>

<pre>
  &lt;action monitor=&quot;[renderobject]&quot; call=&quot;hide&quot;&gt;
    &lt;...&gt;
  &lt;/action&gt;
</pre>

<p>For a complete list of actions and their relevant arguments, you will
need to refer to the Action Reference Guide for more detailed
information.</p>

###ACTIONS###
Activate:         Activates an Action object.
ClosingTag:       If the object is non-static, this action will cause the Action object to activate and then self-destruct.
SetUnlistedField: Unlisted fields are supported for action and method argument values.

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_ACTION 1.1

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_ACTION, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003.  All rights reserved.",
   "March 2003",
   "Action"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static struct ActionTable *ActionTable = 0;
static OBJECTPTR modStrings = 0;
static OBJECTPTR ActionClass = 0;
static OBJECTID glBroadcastID = 0;

/*****************************************************************************
** Class definition.
*/

#define LEN_CALL 40
#define LEN_OBJECTREFERENCE 100
#define SIZE_ARGS 100
#define MAX_FIELDS 16

enum {
   TYPE_ACTION=1,
   TYPE_METHOD
};

struct Action {
   OBJECT_HEADER        /* Class header */
   LONG     Static;     /* Set to TRUE if the action object should remain in memory */
   ERROR    Error;      /* Reflects the error condition of the last executed action */
   LONG     Flags;      /* Optional flags */

   PRIVATE_FIELDS
   OBJECTID ObjectID;
   OBJECTID ClassID;          /* The class of the object in the ObjectID field */
   ACTIONID ActionID;         /* The related action or method ID */
   BYTE     ObjectReference[LEN_OBJECTREFERENCE];
   BYTE     Call[LEN_CALL];   /* The action or method to call */
   BYTE     Arguments[SIZE_ARGS];
   WORD     Type;             /* Either TYPE_METHOD or TYPE_ACTION */
   WORD     ArgsSize;
   WORD     Monitor;
   STRING   Fields[MAX_FIELDS+1];
   UBYTE    ArgsSet;         /* Count of arguments that have been set */
};

static ERROR GET_Call(struct Action *, STRING *);
static ERROR GET_Custom(struct Action *, STRING *);
static ERROR GET_Monitor(struct Action *, OBJECTID *);
static ERROR GET_Object(struct Action *, struct Variable *);

static ERROR SET_Call(struct Action *, STRING);
static ERROR SET_Custom(struct Action *, STRING);
static ERROR SET_Method(struct Action *, STRING);
static ERROR SET_Monitor(struct Action *, OBJECTID *);
static ERROR SET_Object(struct Action *, struct Variable *);

#define AF_TRANSLATE 0x00000001
#define AF_LOCAL     0x00000002
#define AF_MESSAGE   0x00000004
#define AF_NOMSG     AF_LOCAL

struct FieldDef ActionFlags[] = {
   { "TRANSLATE", AF_TRANSLATE },
   { "LOCAL",     AF_LOCAL },
   { "MESSAGE",   AF_MESSAGE },
   /*** Synonyms ***/
   { "NOMSG",     AF_NOMSG },
   { NULL, NULL }
};

static struct FieldArray ActionFields[] = {
   { "Static", 0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Error",  0, FDF_LONG|FDF_R,       0, NULL, NULL },
   { "Flags",  0, FDF_LONGFLAGS|FDF_RI, (LONG)&ActionFlags, NULL, NULL },
   /*** Virtual fields ***/
   { "Monitor", 0, FDF_OBJECTID|FDF_RW, 0, GET_Monitor, SET_Monitor },
   { "Call",    0, FDF_STRING|FDF_RW,   0, GET_Call,    SET_Call },
   { "Custom",  0, FDF_STRING|FDF_RW,   0, GET_Custom,  SET_Custom },
   { "Method",  0, FDF_STRING|FDF_RW,   0, GET_Call,    SET_Method },
   { "Object",  0, FDF_VARIABLE|FDF_STRING|FDF_RW, 0, GET_Object, SET_Object },
   END_FIELD
};

static ERROR ACTION_ActionNotify(struct Action *, struct acActionNotify *);
static ERROR ACTION_Activate(struct Action *, APTR);
static ERROR ACTION_ClosingTag(struct Action *, APTR);
static ERROR ACTION_EventMessage(struct Action *, struct acEventMessage *);
static ERROR ACTION_Free(struct Action *, APTR);
static ERROR ACTION_GetUnlistedField(struct Action *, struct acGetUnlistedField *);
static ERROR ACTION_Init(struct Action *, APTR);
static ERROR ACTION_NewObject(struct Action *, APTR);
static ERROR ACTION_SetUnlistedField(struct Action *, struct acSetUnlistedField *);

static struct ActionArray ActionActions[] = {
   { AC_ActionNotify,     ACTION_ActionNotify },
   { AC_Activate,         ACTION_Activate },
   { AC_ClosingTag,       ACTION_ClosingTag },
   { AC_EventMessage,     ACTION_EventMessage },
   { AC_Free,             ACTION_Free },
   { AC_GetUnlistedField, ACTION_GetUnlistedField },
   { AC_Init,             ACTION_Init },
   { AC_NewObject,        ACTION_NewObject },
   { AC_SetUnlistedField, ACTION_SetUnlistedField },
   { NULL, NULL }
};

static ERROR SetArgument(struct Action *Self, STRING Field, STRING Value);

/*****************************************************************************
** User Instructions.
*/

static BYTE ActionUsage[] = {
"USAGE: action object=name|id [call|method]=action [args...]\n\
\n\
You can call any action or method on an object by using this\n\
command.  This command is considered to be very powerful, so we\n\
recommend that you read the Action List manual in the Pandora\n\
Engine SDK before you use it.\n\
\n\
Examples\n\
--------\n\
If you wanted to destroy an object that you knew the ID of,\n\
you could use this command:\n\
\n\
  action object=-5329 call=free\n\
\n\
If you wanted to move a particular window to the front of the\n\
display:\n\
\n\
  action object=mywindow call=movetofront\n"
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /*** Get the global action table from the kernel ***/

   if (!(ActionTable = ActionList())) {
      return(ObjectError(argModule, ERH_InitModule, ERR_Failed));
   }

   /*** Find the SystemBroadcaster ***/

   FastFindObject("SystemBroadcaster", ID_BROADCAST, &glBroadcastID, 1, NULL);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ActionClass, NULL,
      FID_BaseClassID|TLONG, ID_ACTION,
      FID_Version|TFLOAT,    VER_ACTION,
      FID_Name|TSTRING,      "Action",
      FID_Category|TLONG,    CCF_COMMAND,
      FID_Usage|TPTR,        ActionUsage,
      FID_Actions|TPTR,      ActionActions,
      FID_Fields|TPTR,       ActionFields,
      FID_Size|TLONG,        sizeof(struct Action),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (modStrings)  { acFree(modStrings);  modStrings  = NULL; }
   if (ActionClass) { acFree(ActionClass); ActionClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ActivateObject()
**
** We use Action() only if we are operating on a private object or if AF_LOCAL
** is specified.  Note that ActionMsg() returns ERR_Okay even if the action
** itself fails, so the use of AF_LOCAL is useful when you need the error code.
*/

static ERROR ActivateObject(struct Action *Self, ACTIONID ActionID, OBJECTID ObjectID)
{
   OBJECTPTR object;
   ERROR error;

   if ((ObjectID > 0) OR (Self->Flags & AF_LOCAL)) {
      if (AccessObject(ObjectID, 5000, &object) IS ERR_Okay) {
         if (Self->ArgsSize) error = Action(ActionID, object, Self->Arguments);
         else error = Action(ActionID, object, NULL);
         ReleaseObject(object);
      }
      else error = ERR_AccessObject;
   }
   else {
      if (Self->Flags & AF_MESSAGE) {
         if ((Self->ArgsSize) AND (Self->ArgsSet)) error = DelayMsg(ActionID, ObjectID, Self->Arguments);
         else error = DelayMsg(ActionID, Self->ObjectID, NULL);
      }
      else {
         if ((Self->ArgsSize) AND (Self->ArgsSet)) error = ActionMsg(ActionID, ObjectID, Self->Arguments);
         else error = ActionMsg(ActionID, ObjectID, NULL);
      }
   }

   return(error);
}

/*****************************************************************************
** Action: ActionNotify
*/

static ERROR ACTION_ActionNotify(struct Action *Self, struct acActionNotify *Args)
{
   LONG i, count;

   if (Self->ActionID IS Args->ActionID) {
      /* If we are waiting on a custom call, make sure that the custom string
      ** matches what we are expecting.  NB: This only works when monitoring
      ** custom actions reported from the same task space, because pointers
      ** reported via ActionNotify are always invalid when sent between tasks.
      */

      if (Self->ActionID IS AC_Custom) {
         struct acCustom *custom;
         custom = (struct acCustom *)Args->Args;
         if (!custom) return(ObjectError(Self, ERH_ActionNotify, ERR_Args));

         if (StrCompare(custom->String, ((struct acCustom *)Self->Arguments)->String, 0, STR_MATCHLENGTH) != ERR_Okay) {
            //DPrintF("Action:","Received custom call, but string '%s' != '%s'", custom->String, ((struct acCustom *)Self->Arguments)->String);
            return(ERR_Okay);
         }
      }

      /*** Activate the children ***/

      if ((count = TotalChildren(Self->Head.UniqueID)) > 0) {
         struct ChildEntry list[count];
         if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
            for (i=0; i < count; i++) {
               ActivateObject(Self, AC_Activate, list[i].ObjectID);
               //ActionMsg(AC_Activate, list[i].ObjectID, NULL);
            }
         }
      }
      else DPrintF("ActionNotify:","[Action:%d] I have no child objects that need to be activated.", Self->Head.UniqueID);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Activate
*/

static ERROR ACTION_Activate(struct Action *Self, APTR Void)
{
   STRING fieldvalue;
   WORD i;
   BYTE str[1024], tmp[1024];

   if (Self->ActionID IS AC_Custom) {
      DPrintF("Activate()","[Action:%d] Custom: %s, Object: %d", Self->Head.UniqueID, ((struct acCustom *)Self->Arguments)->String, Self->ObjectID);
   }
   else DPrintF("Activate()","[Action:%d] Call: %s, Object: %d", Self->Head.UniqueID, Self->Call, Self->ObjectID);

   if (Self->ObjectReference[0]) {
      if (StrCopy(Self->ObjectReference, str, sizeof(str)) < sizeof(str)) {
         if (StrTranslate(str, sizeof(str), NULL) IS ERR_Okay) Self->ObjectID = (OBJECTID)StrToInt(str);
         else Self->ObjectID = (OBJECTID)StrToInt(Self->ObjectReference);
      }
      else return((Self->Error = ObjectError(Self, ERH_Activate, ERR_BufferOverflow)));
   }

   if (!Self->ActionID) {
      DPrintF("@Activate:","[Action] Action field not set.");
      return(Self->Error = ERR_FieldNotSet);
   }

   if (!Self->ObjectID) {
      DPrintF("@Activate:","[Action] Object field not set.");
      return(Self->Error = ERR_FieldNotSet);
   }

   if (Self->Flags & AF_TRANSLATE) {
      for (i=0; Self->Fields[i]; i++) {
         for (fieldvalue=Self->Fields[i]; *fieldvalue; fieldvalue++);
         fieldvalue++;

         StrCopy(fieldvalue, str, sizeof(str));
         StrTranslate(str, sizeof(str), NULL);

         if ((*str IS '=') AND (StrCalculate(str+1, NULL, tmp, sizeof(tmp)) IS ERR_Okay)) {
            StrCopy(tmp, str, sizeof(str));
         }

         SetArgument(Self, Self->Fields[i], str);
      }
   }

   /*** Activate the target object here ***/

   Self->Error = ActivateObject(Self, Self->ActionID, Self->ObjectID);

   if (Self->Error) ObjectError(Self, ERH_Activate, Self->Error);

   return(Self->Error);
}

/*****************************************************************************
** Action: ClosingTag
*/

static ERROR ACTION_ClosingTag(struct Action *Self, APTR Void)
{
   if (Self->Monitor) return(ERR_Okay);

   if (Self->Static IS FALSE) {
      /*** Unless we are static, Activate and then Free ourselves ***/
      Action(AC_Activate, Self, NULL);
      Action(AC_Free, Self, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Action: EventMessage
*/

static ERROR ACTION_EventMessage(struct Action *Self, struct acEventMessage *Args)
{
   LONG count;
   WORD i;

   if ((Args->EventID IS EVT_TaskCrashed) AND (Self->ActionID IS AC_Free)) {
      DPrintF("EventMessage()","[Action:%d] Task crash detected.", Self->Head.UniqueID);

      if (CheckObjectExists(Self->ObjectID, NULL) != ERR_Okay) {
         if ((count = TotalChildren(Self->Head.UniqueID)) > 0) {
            struct ChildEntry list[count];
            if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
               for (i=0; i < count; i++) {
                  ActionMsg(AC_Activate, list[i].ObjectID, NULL);
               }
            }
         }
         else DPrintF("ActionNotify:","[Action:%d] I have no child objects that need to be activated.", Self->Head.UniqueID);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Free
*/

static ERROR ACTION_Free(struct Action *Self, APTR Void)
{
   struct MemInfo info;
   struct FunctionField *function;
   struct MethodArray *methods;
   OBJECTPTR class, broadcast;
   LONG offset;
   APTR memory;
   WORD i;

   if ((Self->Monitor) AND (Self->ActionID IS AC_Free)) {
      if (AccessObject(glBroadcastID, 5000, &broadcast) IS ERR_Okay) {
         ActionTags(MT_UnsubscribeEvent, broadcast, Self->Head.UniqueID, EVT_TaskCrashed);
         ReleaseObject(broadcast);
      }
   }

   for (i=0; Self->Fields[i]; i++) {
      FreeMemory(Self->Fields[i]);
      Self->Fields[i] = NULL;
   }

   if (Self->ArgsSize) {
      /*** If any strings were allocated in the argument structure, we must free them ***/

      if (Self->ActionID) {
         function = NULL;
         if (Self->ActionID < 0) {
            if (Self->ClassID) {
               if ((class = FindClass(Self->ClassID, NULL))) {
                  if ((GetField(class, FID_Methods, FT_POINTER, &methods) IS ERR_Okay) AND (methods)) {
                     function = methods[-Self->ActionID].Args;
                  }
               }
            }
         }
         else function = ActionTable[Self->ActionID].Args;

         if (function) {
            offset = 0;
            while (function->Name) {
               if (function->Type & ARG_STRING) {
                  if ((memory = ((STRING *)(Self->Arguments+offset))[0])) {
                     if (MemoryPtrInfo(memory, VER_MemoryInfo, &info) IS ERR_Okay) {
                        ReleaseMemory(memory);
                        FreeMemoryID(info.MemoryID);
                     }
                  }
               }

               if (function->Type & ARG_LONG) offset += sizeof(LONG);
               else if (function->Type & ARG_LARGE) offset += sizeof(LARGE);
               else if (function->Type & (ARG_PTR|ARG_STRING)) offset += sizeof(APTR);
               else if (function->Type & ARG_DOUBLE) offset += sizeof(DOUBLE);
               function++;
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: This action is supported for retrieving unlisted field values.
###END###

*****************************************************************************/

static ERROR ACTION_GetUnlistedField(struct Action *Self, struct acGetUnlistedField *Args)
{
   STRING fieldvalue;
   LONG i, j;

   for (i=0; Self->Fields[i]; i++) {
      if (StrCompare(Args->Field, Self->Fields[i], NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         for (fieldvalue=Self->Fields[i]; *fieldvalue; fieldvalue++);
         fieldvalue++;

         for (j=0; (fieldvalue[j]) AND (j < Args->Size-1); j++) Args->Buffer[j] = fieldvalue[j];
         Args->Buffer[j++] = 0;

         if (j >= Args->Size) return(ERR_BufferOverflow);
         else return(ERR_Okay);
      }
   }

   DPrintF("@GetVariable:","[Set] The variable \"%s\" does not exist.", Args->Field);

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Init
*/

static ERROR ACTION_Init(struct Action *Self, APTR Void)
{
   OBJECTPTR monitor, broadcast;

   if (!Self->ObjectID) Self->ObjectID = Self->Head.OwnerID;

   if (!Self->ActionID) {
      if ((Self->Type IS TYPE_METHOD) AND (Self->ObjectID)) {
         SetField(Self, FID_Method, FT_STRING, Self->Call);

         if (!Self->ActionID) {
            DPrintF("@Init:","[Action] Unable to resolve method \"%s\" for object %d", Self->Call, Self->ObjectID);
            return(ERR_FieldNotSet);
         }

      }
      else {
         if ((!Self->ActionID) AND (!Self->ObjectID)) DPrintF("@Init:","[Action:%d] Missing both the action and object fields.", Self->Head.UniqueID);
         else if (Self->ActionID) DPrintF("@Init:","[Action:%d] Missing the object field (call %s).", Self->Head.UniqueID, Self->Call);
         else DPrintF("@Init:","[Action:%d] Missing the action field (object #%d).", Self->Head.UniqueID, Self->ObjectID);
         return(ERR_FieldNotSet);
      }
   }

   /* If action monitoring is to be used, subscribe
   ** to the action here.
   */

   if (Self->Monitor) {
      if (AccessObject(Self->ObjectID, 5000, &monitor) IS ERR_Okay) {
         SubscribeAction(monitor, Self->ActionID, Self);
         ReleaseObject(monitor);
      }

      /* When the Free action is being monitored, subscribed to the TaskCrashed
      ** event so that we know if the monitored object is ever destroyed by a
      ** crash.
      */

      if (Self->ActionID IS AC_Free) {
         if (AccessObject(glBroadcastID, 5000, &broadcast) IS ERR_Okay) {
            ActionTags(MT_SubscribeEvent, broadcast, Self->Head.UniqueID, EVT_TaskCrashed);
            ReleaseObject(broadcast);
         }
      }
   }

   Self->Error = ERR_Failed;

   return(ERR_Okay);
}

/*****************************************************************************
** Action: SetUnlistedField
*/

static ERROR ACTION_SetUnlistedField(struct Action *Self, struct acSetUnlistedField *Args)
{
   STRING fieldname;
   LONG len;
   WORD i;

   if ((!Args) OR (!Args->Field) OR (!Args->Value)) {
      return(ObjectError(Self, ERH_SetUnlistedField, ERR_Args));
   }

   if (Self->Flags & AF_TRANSLATE) {
      /* If translation mode is enabled, store the argument name and value for
      ** future translation.
      */

      for (i=0; Self->Fields[i]; i++) {
         if (StrCompare(Args->Field, Self->Fields[i], NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            FreeMemory(Self->Fields[i]);
            Self->Fields[i] = NULL;
            break;
         }
      }

      if (i < MAX_FIELDS) {
         if (AllocMemory(StrLength(Args->Field) + StrLength(Args->Value) + 2,
               MEM_STRING|MEM_NOCLEAR, (void **)&fieldname, NULL) IS ERR_Okay) {

            Self->Fields[i] = fieldname;

            if (Args->Field[0] IS '?') i = StrCopy(Args->Field+1, fieldname, COPY_ALL) + 1;
            else i = StrCopy(Args->Field, fieldname, COPY_ALL) + 1;

            StrCopy(Args->Value, fieldname + i, COPY_ALL);

            if (fieldname[i] IS '<') {
               len = StrLength(fieldname+i);
               if (fieldname[i+len-1] IS '>') {
                  fieldname[i] = '[';
                  fieldname[i+len-1] = ']';
               }
            }
         }
      }
      else return(ERR_Failed);

      return(ERR_Okay);
   }

   if (!Self->ActionID) {
      /*** Resolve the action/method ID if we have not been able to do so already ***/

      if (!Self->ObjectID) {
         Self->ObjectID = Self->Head.OwnerID;
         Self->ClassID  = GetClassID(Self->ObjectID);
         Self->ObjectReference[0] = 0;
      }

      if (Self->Type IS TYPE_METHOD) {
         if (Self->ObjectID) {
            SetField(Self, FID_Method, FT_POINTER, Self->Call);
         }
         else {
            DPrintF("@SetField()","[Action] You need to set an object before setting field \"%s\".", Args->Field);
            return(ERR_FieldNotSet);
         }
      }

      if (!Self->ActionID) {
         DPrintF("@SetUnlistedField()","[Action] You need to give me an action or method before setting field \"%s\".", Args->Field);
         return(ERR_FieldNotSet);
      }
   }
   else if (!Self->ArgsSize) {
      DPrintF("@SetUnlistedField()","[Action] The %s action does not support arguments (%s).", Self->Call, Args->Field);
      return(ERR_FieldNotSet);
   }

   SetArgument(Self, Args->Field, Args->Value);

   return(ERR_Okay);
}

static ERROR SetArgument(struct Action *Self, STRING Field, STRING Value)
{
   struct FunctionField *field;
   struct MethodArray *methods;
   OBJECTPTR class;
   OBJECTID ObjectID;
   STRING string;
   LONG offset, size, len;

   field = NULL;
   size  = NULL;

   if (Self->ActionID < 0) {
      /*** Method argument handling (requires that we lookup the class) ***/

      if ((!Self->ObjectID) OR (!Self->ClassID)) {
         DPrintF("!SetUnlistedField()","[Action:%d] You need to set the Object field before defining arguments.", Self->Head.UniqueID);
         return(ERR_FieldNotSet);
      }

      if ((class = FindClass(Self->ClassID, NULL))) {
         if ((GetField(class, FID_Methods, FT_POINTER, &methods) IS ERR_Okay) AND (methods)) {
            field = methods[-Self->ActionID].Args;
            size = methods[-Self->ActionID].StructureSize;
         }
         else return(ObjectError(Self, ERH_SetUnlistedField, ERR_NoMethods));
      }
      else return(ObjectError(Self, ERH_SetUnlistedField, ERR_MissingClass));
   }
   else {
      /*** Action argument handling ***/
      field = ActionTable[Self->ActionID].Args;
      size = ActionTable[Self->ActionID].Size;
   }

   /*** Find the argument and set it ***/

   offset = 0;
   while (field->Name) {
      if (StrCompare(field->Name, Field, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         if (field->Type & ARG_LONG) {
            if (field->Type & ARG_OBJECT) {
               if (FastFindObject(Value, NULL, &ObjectID, 1, NULL) IS ERR_Okay) {
                  ((LONG *)(Self->Arguments+offset))[0] = ObjectID;
               }
            }
            else ((LONG *)(Self->Arguments+offset))[0] = (LONG)StrToInt(Value);
         }
         else if (field->Type & ARG_LARGE) ((LARGE *)(Self->Arguments+offset))[0] = StrToInt(Value);
         else if (field->Type & ARG_STRING) {
            /*** Free any existing string setting first ***/

            if ((string = ((STRING *)(Self->Arguments+offset))[0])) {
               FreeMemory(string);
               ((STRING *)(Self->Arguments+offset))[0] = NULL;
            }

            for (len=0; Value[len]; len++);
            if (AllocMemory(len+1, MEM_STRING, (void **)&string, NULL) IS ERR_Okay) {
               for (len=0; Value[len]; len++) string[len] = Value[len];
               string[len] = 0;
               ((STRING *)(Self->Arguments+offset))[0] = string;
            }
         }
         else if (field->Type & ARG_PTR)    ((APTR *)(Self->Arguments+offset))[0]   = (APTR)(LONG)StrToInt(Value);
         else if (field->Type & ARG_DOUBLE) ((DOUBLE *)(Self->Arguments+offset))[0] = StrToFloat(Value);
         else DPrintF("@SetUnlistedField:","[Action] Unknown argument definition %s / %d.", Self->Call, field->Type);
         break;
      }

      if (field->Type & ARG_LONG)        offset += sizeof(LONG);
      else if (field->Type & ARG_LARGE)  offset += sizeof(LARGE);
      else if (field->Type & ARG_PTR)    offset += sizeof(APTR);
      else if (field->Type & ARG_STRING) offset += sizeof(STRING);
      else if (field->Type & ARG_DOUBLE) offset += sizeof(DOUBLE);
      else DPrintF("@SetUnlistedField:","[Action] Bad argument definition found for action \"%s\", argument \"%s\"", Self->Call, field->Name);

      field++;
      if (offset >= size) break;
   }

   /*** If we got to the end of the list, then the argument does not exist ***/

   if (!field->Name) DPrintF("@SetUnlistedField:","[Action] Unknown argument \"%s\" for action \"%s\".", Field, Self->Call);
   else Self->ArgsSet = TRUE;

   return(ERR_Okay);
}

/*****************************************************************************
** Action: NewObject
*/

static ERROR ACTION_NewObject(struct Action *Self, APTR Void)
{
   Self->Static = FALSE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Call
Short:  Set this field to specify the action that you want to call.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>This field (or the method field) must be set to determine the action that
will be executed or monitored by an action object.  A list of valid action
names are provided in the manual describing the available system actions.</p>

###NOTE###
<p>It is vital that the correct arguments are provided with the action that
you are going to call - failure to set the arguments can give unpredictable
results.</p>

###SEE ALSO###
Field: Method

###END###

*****************************************************************************/

static ERROR GET_Call(struct Action *Self, STRING *Value)
{
   *Value = Self->Call;
   return(ERR_Okay);
}

static ERROR SET_Call(struct Action *Self, STRING Value)
{
   struct MethodArray *MethodTable;
   OBJECTPTR class;
   LONG totalmethods;
   WORD i;

   Self->ActionID = NULL;

   if ((!Value) OR (!*Value)) {
      Self->Call[0] = 0;
      return(ERR_Okay);
   }

   for (i=0; (Value[i]) AND (i < LEN_CALL-1); i++) Self->Call[i] = Value[i];
   Self->Call[i] = 0;

   for (i=1; ActionTable[i].Name; i++) {
      if (StrCompare(ActionTable[i].Name, Self->Call, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->ArgsSize = ActionTable[i].Size;
         Self->ActionID = i;
         Self->Type     = TYPE_ACTION;
         return(ERR_Okay);
      }
   }

   /*** Test if the reference is to a method (only possible if we know the target object) ***/

   if ((Self->ObjectID) AND (Self->ClassID)) {
      if ((class = FindClass(Self->ClassID, NULL))) {
         if ((GetFields(class, FID_Methods|TPTR,       &MethodTable,
                               FID_TotalMethods|TLONG, &totalmethods,
                               TAGEND) IS ERR_Okay) AND (MethodTable)) {
            for (i=1; i < totalmethods+1; i++) {
               if (StrCompare(Value, MethodTable[i].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  Self->ArgsSize = MethodTable[i].StructureSize;
                  Self->ActionID = MethodTable[i].MethodID;
                  Self->Type     = TYPE_METHOD;
                  return(ERR_Okay);
               }
            }
         }
      }
   }

   /* Consider the setting to be to a customised action if it is not
   ** a valid action or method.
   */

   DPrintF("Set.Call:","[Action:%d] Action \"%s\" not recognised - assuming customised.", Self->Head.UniqueID, Self->Call);

   SET_Custom(Self, Value);

   return(ERR_Search);
}

/******************************************************************************

###FIELD###
Name:  Custom
Short: The Custom field provides a customised action service.
Type:  STRING

###DESCRIPTION###
<p>The Custom field is used in circumstances where you want to execute a custom
action on an object, or to create a customised action for an
existing object.   When setting the Custom field, you need to provide a unique
name that does not match any of the existing action names (it is recommended
that you steer clear of using generic terms such as 'Refresh').  If you set
the Custom field, you must not set it in conjunction with the Method or Call
fields, or you will lose your definition.  Also, you may not pass your own
arguments via custom actions, although you may set the 'Number' argument
that accompanies the customised action standard.</p>

###END###

******************************************************************************/

static ERROR GET_Custom(struct Action *Self, STRING *Value)
{
   *Value = Self->Call;
   return(ERR_Okay);
}

static ERROR SET_Custom(struct Action *Self, STRING Value)
{
   WORD i;

   if ((!Value) OR (!*Value)) return(ERR_NoData);

   for (i=0; "Custom"[i]; i++) Self->Call[i] = "Custom"[i];
   Self->Call[i] = 0;

   ((struct acCustom *)(Self->Arguments))->Number = 0;
   ((struct acCustom *)(Self->Arguments))->String = StrClone(Value);
   Self->ArgsSize = sizeof(struct acCustom);
   Self->ArgsSet  = TRUE;
   Self->ActionID = AC_Custom;
   Self->Type     = TYPE_ACTION;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:    Error
Short:   Reflects the error code of the last executed action.
Type:    LONG
Status:  Read
Version: 1.1

###DESCRIPTION###
<p>When an Action object is activated, the Error field will be updated to
reflect the error code that was returned from that action or method.
A string describing the error code can be loaded from the
'system:config/error_messages.cfg' config file.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Method
Short:  Set if you want to execute a method rather than an action.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>If a method should be executed rather than an action, set this field
to the name of the method that you want to execute.  If the method requires
that a certain number of arguments need to be set, you <i>must also set those
arguments</i> via the unlisted field mechanism, to ensure that the method call
is legal.</p>

###SEE ALSO###
Field: Call

###END###

*****************************************************************************/

static ERROR SET_Method(struct Action *Self, STRING Method)
{
   struct MethodArray *MethodTable;
   OBJECTPTR class;
   LONG i, TotalMethods;

   if (!Method) {
      Self->Call[0] = 0;
      return(ERR_Okay);
   }

   for (i=0; (Method[i]) AND (i < sizeof(Self->Call)); i++) Self->Call[i] = Method[i];
   Self->Call[i] = 0;

   if (!Self->ObjectID) {
      Self->Type = TYPE_METHOD;
      return(ERR_Okay);
   }

   if (Self->ClassID) {
      if ((class = FindClass(Self->ClassID, NULL))) {
         if ((GetFields(class, FID_Methods|TPTR,       &MethodTable,
                               FID_TotalMethods|TLONG, &TotalMethods,
                               TAGEND) IS ERR_Okay) AND (MethodTable)) {
            for (i=1; i < TotalMethods+1; i++) {
               if (StrCompare(Method, MethodTable[i].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  Self->ArgsSize = MethodTable[i].StructureSize;
                  Self->ActionID = MethodTable[i].MethodID;
                  Self->Type     = TYPE_METHOD;
                  return(ERR_Okay);
               }
            }
            DPrintF("@Set.Method:","[Action] Could not find method \"%s\" from %d methods in class %d.", Method, TotalMethods, Self->ClassID);
            return(ERR_Search);
         }
         else {
            DPrintF("@Set.Method:","[Action] Object #%d does not support any methods (class %d).", Self->ObjectID, Self->ClassID);
            return(ERR_NoMethods);
         }
      }
      else return(ObjectError(Self, ERH_SetField, ERR_Search));
   }
   else return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Monitor
Short:  If you want to monitor an object for activity, set this field.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>By setting the Monitor field to an existing ObjectID, your action object
will only react when the monitored object completes a specific action.
Because monitoring requires that the action object is static, setting the
Monitor field will automatically ensure that the action object is not
freed during script execution.</p>

<p>In the following example, action monitoring is used to react to an
object when it is freed:</p>

<pre>
   &lt;action monitor="[myobject]" call="free"/&gt;
      &lt;action static call="hide" object="[{name}]"/&gt;
   &lt/action&gt;
</pre>

###SEE ALSO###
Field: Call, Object

###END###

*****************************************************************************/

static ERROR GET_Monitor(struct Action *Self, OBJECTID *ObjectID)
{
   *ObjectID = Self->ObjectID;
   return(ERR_Okay);
}

static ERROR SET_Monitor(struct Action *Self, OBJECTID *ObjectID)
{
   Self->Monitor  = TRUE;
   Self->Static   = TRUE;
   Self->ObjectID = *ObjectID;
   Self->ClassID  = GetClassID(Self->ObjectID);
   Self->ObjectReference[0] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Object
Short:  Refers to the object that will receive the action.
Type:   STRING/OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>The object that is to receive the action call must be specified in this
field.  You are required to set this field unless you intend to use action
monitoring, in which case you should set the Monitor field.</p>

<p>If you set this field with a string, the object reference will be
translated to an ID in run-time.</p>

###SEE ALSO###
Fields: Call, Monitor

###END###

*****************************************************************************/

static ERROR GET_Object(struct Action *Self, struct Variable *Value)
{
   BYTE buffer[200];

   if (Value->Type & FD_LARGE) {
      if (Self->ObjectReference[0]) {
         if (StrCopy(Self->ObjectReference, buffer, sizeof(buffer)) < sizeof(buffer)) {
            if (StrTranslate(buffer, sizeof(buffer), NULL) IS ERR_Okay) {
               Value->Large = StrToInt(buffer);
            }
            else Value->Large = StrToInt(Self->ObjectReference);
         }
         else return(ObjectError(Self, ERH_SetField, ERR_BufferOverflow));
      }
      else Value->Large = Self->ObjectID;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
}

static ERROR SET_Object(struct Action *Self, struct Variable *Value)
{
   BYTE buffer[200];
   LONG i;

   if (Value->Type & FD_LARGE) {
      Self->ObjectID = (OBJECTID)Value->Large;
      Self->ClassID  = GetClassID(Self->ObjectID);
      Self->ObjectReference[0] = 0;
      return(ERR_Okay);
   }
   else if (Value->Type & FD_STRING) {
      if ((!Value->Pointer) OR (!((STRING)Value->Pointer)[0])) return(ERR_Okay);

      for (i=0; (((STRING)Value->Pointer)[i]) AND (((STRING)Value->Pointer)[i] != '['); i++);

      if (((STRING)Value->Pointer)[i] IS '[') {
         if (StrCopy(Value->Pointer, Self->ObjectReference, LEN_OBJECTREFERENCE) < LEN_OBJECTREFERENCE) {
            if (StrCopy(Self->ObjectReference, buffer, sizeof(buffer)) < sizeof(buffer)) {
               if (StrTranslate(buffer, sizeof(buffer), NULL) IS ERR_Okay) Self->ObjectID = (OBJECTID)StrToInt(buffer);
               else Self->ObjectID = (OBJECTID)StrToInt(Self->ObjectReference);
               Self->ClassID = GetClassID(Self->ObjectID);
               return(ERR_Okay);
            }
            else return(ObjectError(Self, ERH_SetField, ERR_BufferOverflow));
         }
         else return(ObjectError(Self, ERH_SetField, ERR_BufferOverflow));
      }
      else {
         FastFindObject(Value->Pointer, NULL, &Self->ObjectID, 1, NULL);
         Self->ClassID = GetClassID(Self->ObjectID);
         return(ERR_Okay);
      }
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));
}

/*****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE if the Action should be static.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>By default, an Action object will execute itself and then self-destruct
when the closing tag is received.  If you would rather that the object stays
in the system, set this field to TRUE.  If you do this, the only way to get
the Action object to perform is to call the Activate() action.</p>

###END###

*****************************************************************************/
